#==========================================================================================#

# VCE TYPES

abstract type CorrStructure
end

mutable struct Homoscedastic <: CorrStructure
    corrected::Bool
    expected::Bool
    nonmissing::Bool
end

mutable struct Heteroscedastic <: CorrStructure
    corrected::Bool
    nonmissing::Bool
end

mutable struct Clustered <: CorrStructure
    corrected::Bool
    nonmissing::BitVector
    matrix::AbstractMatrix
    id_clusters::AbstractVector
    n_clusters::Int
end

mutable struct CrossCorrelated <: CorrStructure
    corrected::Bool
    nonmissing::BitVector
    matrix::AbstractMatrix
end

#==========================================================================================#

# HOMOSCEDASTIC

function Homoscedastic(; corrected::Bool = true, expected::Bool = false)
    Homoscedastic(corrected, expected, true)
end

# HETEROSCEDASTIC

Heteroscedastic(; corrected::Bool = true) = Heteroscedastic(corrected, true)

#==========================================================================================#

# CLUSTERED

function Clustered(df::DataFrame, x::Symbol; corrected::Bool = true)

    nonmissing  = completecases(df, x)
    id_clusters = disallowmissing(df[nonmissing, x])
    iter        = sort(unique(id_clusters))
    n_clusters  = length(iter)
    idx₁        = Vector{Int}()
    idx₂        = Vector{Int}()

    @inbounds for (i, ci) in enumerate(iter)
        idx = findall((in)([ci]), id_clusters)
        append!(idx₁, fill(i, length(idx)))
        append!(idx₂, idx)
    end

    val = fill(1.0, length(idx₁))
    V   = SuiteSparse.CHOLMOD.Sparse(sparse(idx₁, idx₂, val))

    return Clustered(corrected, nonmissing, V, id_clusters, n_clusters)
end

#==========================================================================================#

# CROSS CORRELATED: GENERAL CONSTRUCTOR

function CrossCorrelated(variant::String, args...; kwargs...)
    if variant == "Two-way clustering"
        return cc_twowayclustering(args...; kwargs...)
    elseif variant == "Time"
        return cc_time(args...; kwargs...)
    elseif variant == "Space"
        return cc_space(args...; kwargs...)
    elseif variant == "Time and space"
        return cc_timespace(args...; kwargs...)
    else
        throw("unknown variant")
    end
end

# TWOWAY CLUSTERING

function cc_twowayclustering(df::DataFrame, x₁::Symbol, x₂::Symbol; corrected::Bool = true)

    nonmissing  = completecases(df, [x₁, x₂])
    ic₁         = disallowmissing(df[nonmissing, x₁])
    ic₂         = disallowmissing(df[nonmissing, x₂])
    n           = sum(nonmissing)
    iter₁       = unique(ic₁)
    iter₂       = unique(ic₂)
    idx₁        = Vector{Int}(1:n)
    idx₂        = Vector{Int}(1:n)
    nn          = n

    for (iter, ic) in zip([iter₁, iter₂], [ic₁, ic₂])
        @inbounds for i in iter₁

            iter = findall((in)([i]), ic)
            nix  = length(idx)
            nel  = Int(nix * (nix - 1) / 2)

            append!(idx₁, Vector{Int}(nel))
            append!(idx₂, Vector{Int}(nel))

            for j = 1:nix
                for k = 1:(j - 1)
                    nn += 1
                    idx₁[nn] = idx[k]
                    idx₂[nn] = idx[j]
                end
            end
        end
    end

    val = fill(1.0, length(idx₁))
    V   = SuiteSparse.CHOLMOD.Sparse(Symmetric(sparse(idx₁, idx₂, val, n, n, max)))

    return CrossCorrelated(corrected, nonmissing, V)
end

# CORRELATION ACROSS TIME

function cc_time(
        df::DataFrame,
        x::Symbol,
        b::Real,
        k::Function = parzen;
        corrected::Bool = true
    )

    nonmissing = completecases(df, x)
    xx         = Vector{Date}(df[nonmissing, x])
    n          = sum(nonmissing)
    idx₁       = Vector{Int}(1:n)
    idx₂       = Vector{Int}(1:n)
    val        = fill(1.0, n)

    kernel(z) = k(z / float(b))

    @inbounds for i = 1:n
        for j = 1:(i - 1)
            w = Dates.value(xx[i] - xx[j])
            w = k(w)
            if w > 0.0
                push!(idx₁, j)
                push!(idx₂, i)
                push!(val, w)
            end
        end
    end

    V = Symmetric(sparse(idx₁, idx₂, val))

    return CrossCorrelated(corrected, nonmissing, SuiteSparse.CHOLMOD.Sparse(V))
end

# CORRELATION ACROSS SPACE

function cc_space(
        df::DataFrame,
        y::Symbol,
        x::Symbol,
        b::Real,
        k::Function = parzen;
        corrected::Bool = true
    )

    nonmissing  = completecases(df, [y, x])
    yy          = Vector{Float64}(df[nonmissing, y])
    xx          = Vector{Float64}(df[nonmissing, x])
    n           = sum(nonmissing)
    idx₁        = Vector{Int}(1:n)
    idx₂        = Vector{Int}(1:n)
    val         = fill(1.0, n)

    kernel(z) = k(z / float(b))

    @inbounds for i = 1:n
        for j = 1:(i - 1)
            w = geodistance(yy[i], xx[i], yy[j], xx[j])
            w = kernel(w)
            if w > 0.0
                push!(idx₁, j)
                push!(idx₂, i)
                push!(val, w)
            end
        end
    end

    V = Symmetric(sparse(idx₁, idx₂, val))

    return CrossCorrelated(corrected, nonmissing, SuiteSparse.CHOLMOD.Sparse(V))
end

# CORRELATION ACROSS TIME AND SPACE

function cc_timespace(
        df::DataFrame,
        x₁::Symbol,
        b₁::Real,
        y₂::Symbol,
        x₂::Symbol,
        b₂::Real,
        k::Function = parzen;
        corrected::Bool = true
    )

    nonmissing  = completecases(df, [x₁, y₂, x₂])
    xx₁         = Vector{Date}(df[nonmissing, x₁])
    yy₂         = Vector{Float64}(df[nonmissing, y₂])
    xx₂         = Vector{Float64}(df[nonmissing, x₂])
    n           = sum(nonmissing)
    idx₁        = Vector{Int}(1:n)
    idx₂        = Vector{Int}(1:n)
    val         = fill(1.0, n)

    b₁ = float(b₁)
    b₂ = float(b₂)

    kernel(z₁, z₂) = k(sqrt(abs2(z₁) + abs2(z₂)))

    @inbounds for i = 1:n
        for j = 1:(i - 1)
            w₁ = Dates.value(xx₁[i] - xx₁[j]) / b₁
            if w₁ <= 1.0
                w₂ = geodistance(yy₂[i], xx₂[i], yy₂[j], xx₂[j]) / b₂
                if w₂ <= 1.0
                    push!(idx₁, j)
                    push!(idx₂, i)
                    push!(val, kernel(w₁, w₂))
                end
            end
        end
    end

    V = Symmetric(sparse(idx₁, idx₂, val))

    return CrossCorrelated(corrected, nonmissing, SuiteSparse.CHOLMOD.Sparse(V))
end
